/*--------------------------------------------------------------------
 * $Id: ipa.c,v 1.39 2009/12/22 16:06:31 robert Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "uvspec.h"
#include "ascii.h"
#include "wcloud3d.h"

typedef struct {
  int ix;       /* x-index of the pixel     */
  int iy;       /* y-index of the pixel     */

  int nc;       /* number of cloudy pixels  */
  double tau;   /* total optical depth      */
 
  float ***k;  /* extinction coefficient   */
  float ***g;  /* asymmetry parameter      */
  float ***w;  /* single-scattering albedo */

  int cldproperties;

  int nz;       /* number of layers         */
}  sortarray_struct;


#define EPSILON 1E-6


/* prototypes of internal functions */
static int sortarray_fnc (const void *x1, const void *x2);
static int write_pixel (float *z3D, float ***k3D, float ***g3D, float ***w3D, 
			int *threed, int cldproperties, 
			int ix, int iy, int Nz, int index,
			output_struct *output, int quiet);

int setup_ipa (input_struct input, output_struct *output)
{

  /* Reading of ipa input files etc. is done in setup_wcloud */

  if (!input.wcipa && !input.icipa)  {
    /* no independent pixel */
    output->nipa=1;
    output->ipaweight=calloc (1, sizeof(double));
    output->ipaweight[0] = 1;
  }

  return 0;   /* if o.k. */
}


/*********************************/
/* compare two altitude grids    */
/*********************************/

int compare_levels_ipa_df (float *z1, int n1,
			   float *z2, int n2)
{
  int kc=0;

  if (n1!=n2) {
    fprintf (stderr, "Error, number of layers differing (line %d, function %s in %s)\n", __LINE__, __func__, __FILE__);
    fprintf (stderr, "%d in 3D, %d in 1D\n", n1, n2);
    return -1;
  }

  for (kc=0; kc<=n1; kc++)
    if (fabs (z1[kc]/1000.0 - z2[n1-kc]) > EPSILON * z1[kc]) {
      fprintf (stderr, "Error, difference at level %d, altitude %g vs. %g (line %d, function %s in %s)\n",	
	       kc, z1[kc], z2[n1-kc], __LINE__, __func__, __FILE__);
      return -1;
    }

  return 0;  /* if o.k. */
}


/***********************************************************************************/
/* Function: sortarray_fnc                                                         */
/* Description: Internal function, required by ASCII_sortarray();                  */
/*              see description of standard C library function qsort().            */
/*                                                                                 */
/* Parameters:                                                                     */    
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/***********************************************************************************/

static int sortarray_fnc (const void *x1, const void *x2)
{
  sortarray_struct *c1 = (sortarray_struct *) x1;
  sortarray_struct *c2 = (sortarray_struct *) x2;

  int lev=0;
  double sub;

  if (c1->tau > c2->tau)
    return 1;
  
  if (c1->tau < c2->tau)
    return -1;

  /*****************************************************************/
  /* equal optical depths, need to compare number of cloudy pixels */
  /*****************************************************************/
  
  if (c1->nc > c2->nc)
    return 1;
  
  if (c1->nc < c2->nc)
    return -1;

  /******************************************************************************/
  /* equal optical depths and number of cloudy pixels, need to compare profiles */
  /******************************************************************************/

  sub=0;
  for (lev=0; lev<c1->nz; lev++) {
    
    if (c1->cldproperties != CLD_OPTPROP || c2->cldproperties != CLD_OPTPROP) {
      fprintf (stderr, "** FATAL error in ipa.c:sortarray_fnc():\n");
      fprintf (stderr, "** currently only implemented for cldproperties = %d\n", CLD_OPTPROP);
      exit(1);
    }

    /* compare extinction coefficients */
    if ((sub = c1->k[lev][c1->ix][c1->iy] - c2->k[lev][c2->ix][c2->iy]) != 0)
      break;
    
    /* compare asymmetry parameters */
    if ((sub = c1->g[lev][c1->ix][c1->iy] - c2->g[lev][c2->ix][c2->iy]) != 0)
      break;
    
    /* compare single scattering albedo */
    if ((sub = c1->w[lev][c1->ix][c1->iy] - c2->w[lev][c2->ix][c2->iy]) != 0)
      break;
    
  }  

  if (lev==c1->nz)   /* if reached TOA without finding differing layers */
    return 0;
  
  return (sub>0?-1:1);
}





/**********************************************************/
/* Write one column of a cloud, used by threeDtoIPA().    */ 
/**********************************************************/

static int write_pixel (float *z3D, float ***k3D, float ***g3D, float ***w3D, 
			int *threed, int cldproperties, 
			int ix, int iy, int Nz, int index,
			output_struct *output, int quiet)
{
  int iv=0, lev=0, status=0;

  /* allocate memory for this pixel; allocate memory for */
  /* ALL fields even if they are not required because    */
  /* they may be used by other functions.                */

  switch (cldproperties) {
  case CLD_OPTPROP:

    status = calloc_cld_out (&(output->wcipa[index]), output->wl.nlambda_t, Nz+1, output->wc.optprop.nphamat);
    if (status!=0) {
      fprintf (stderr, "Error %d allocating memory for cld_out_struct\n", status);
      return status;
    }
    
    /* copy 1D cloud (from wc_file) to wcipa[index] */
    status = cp_cld_out (&(output->wcipa[index]), output->wc, 0, 0, quiet);
    if (status!=0) {
      fprintf (stderr, "Error %d copying output->wc to output->wcipa[index]\n", status);
      return status;
    }
    
    
    /* profiles need to be reversed ! */
    for (lev=0; lev<Nz; lev++)
      for (iv=0; iv<output->wl.nlambda_t; iv++) {
	
	/* overwrite only levels where actually a 3D cloud is specified */
	if (threed[lev]) {
	  output->wcipa[index].optprop.dtau[iv][Nz-lev-1] = k3D[lev][ix][iy]*(z3D[lev+1]-z3D[lev]);
	  output->wcipa[index].optprop.g1[iv][Nz-lev-1]  = g3D[lev][ix][iy];
	  output->wcipa[index].optprop.ssa[iv][Nz-lev-1] = w3D[lev][ix][iy];
	}
      }
    break;

  case CLD_EXTREFF:
  case CLD_LWCREFF:
    fprintf (stderr, "Error, only cldproperties = %d is currently implemented for wc_ipa!\n", CLD_OPTPROP);
    return -1;
    break;

  default:
    fprintf (stderr, "Error, unknown cloud properties %d\n", cldproperties);
    return -1;
  }
  
  return 0;  /* if ok */
}





/*****************************************/
/* Write 1D cloud data                   */
/*****************************************/

int threeDtoIPA (float *z3D, float ***lwc3D, float ***ref3D, float ***k3D, float ***g3D, float ***w3D,
                 int *threed, int cldproperties, int Nx, int Ny, int Nz,
                 output_struct *output, int quiet)
{
  int keyc=0, index=0;
  int ix=0, iy=0, lev=0, status=0;

  FILE *cldfile=NULL;
  char cldfilename    [FILENAME_MAX+1] = "ipa.idx";
  char cldcolfilename [FILENAME_MAX+1] = "";

  int sarrayc=0;
  int narray = Nx*Ny;

  sortarray_struct *sarray  = calloc(narray, sizeof(sortarray_struct));
  sortarray_struct *sarray1 = calloc(1, sizeof(sortarray_struct));

  /* ?????????? */
  fprintf (stderr, "** Sorry, 3D cloud files in IPA mode are currently not supported!\n");
  fprintf (stderr, "** First need to adapt threeDtoIPA to the slightly changed MYSTIC structures.\n");
  return -1;


  sarrayc=0;
  for (ix=0; ix<Nx; ix++)
    for (iy=0; iy<Ny; iy++) {
      sarray[sarrayc].nz = Nz;
      sarray[sarrayc].ix = ix;
      sarray[sarrayc].iy = iy;
      sarray[sarrayc].k  = k3D;
      sarray[sarrayc].g  = g3D;
      sarray[sarrayc].w  = w3D;

      sarray[sarrayc].cldproperties = cldproperties;

      /* total vertical optical depth */
      sarray[sarrayc].tau   = 0;
      for (lev=0; lev<Nz; lev++)
	sarray[sarrayc].tau += k3D[lev][ix][iy] * (z3D[lev+1] - z3D[lev]);

      /* cloudy pixel counter */
      sarray[sarrayc].nc=0;
      for (lev=0; lev<Nz; lev++)
	if (k3D[lev][ix][iy]>0)
	  sarray[sarrayc].nc++;

      sarrayc++;
    }
  
  qsort (sarray, narray, sizeof(sortarray_struct), sortarray_fnc);
  
  /* create filename for index file */
  if ((cldfile = fopen (cldfilename, "w"))==NULL) {
    fprintf (stderr, "Error opening %s for writing\n", cldfilename);
    return -1;
  }

  fprintf (cldfile, "# Cloud optical depth distribution, generated by %s\n", "uvspec");
  fprintf (cldfile, "# File number, filename, number of cloudy pixels, optical depth, number of occurrences\n");


  sarrayc = 0;
  memcpy (sarray1, &(sarray[0]), sizeof(sortarray_struct)); 
  keyc = 0;
  index=0;

  for (sarrayc=0; sarrayc<narray; sarrayc++) {
    
    /* if different column */
    if (sortarray_fnc (sarray1, &(sarray[sarrayc])) != 0) {
      
      /* write one pixel */
      status = write_pixel (z3D, k3D, g3D, w3D, threed, cldproperties,
			    (sarray[sarrayc-1]).ix, (sarray[sarrayc-1]).iy, 
			    Nz, index, output, quiet);

      /* entry in index file */

      fprintf (cldfile, "%d %s %d %f %d\n", 
	       index, cldcolfilename, 
	       (sarray[sarrayc-1]).nc, 
	       (sarray[sarrayc-1]).tau,
	       keyc);
      
      output->ipaweight[index] = (double) keyc / (double) narray;

      memcpy (sarray1, &(sarray[sarrayc]), sizeof(sortarray_struct)); 
      keyc=1;
      index++;
    }
    else 
      keyc++;
  }
  
  /* write last pixel */
  status = write_pixel (z3D, k3D, g3D, w3D, threed, cldproperties,
			(sarray[sarrayc-1]).ix, (sarray[sarrayc-1]).iy, 
			Nz, index, output, quiet);


  /* last entry in index file */
  fprintf (cldfile, "%d %s %d %f %d\n", 
	   index, cldcolfilename, 
	   (sarray[sarrayc-1]).nc, 
	   (sarray[sarrayc-1]).tau,
	   keyc);

  output->ipaweight[index] = (double) keyc / (double) narray;

  fclose (cldfile);

  output->nipa=index+1;
  
  free(sarray);
  free(sarray1);
  
  return 0;
}



int read_wc_ipa (input_struct input, output_struct *output)
{
  int ipa=0, min_columns=0, max_columns=0, max_length=0, status=0, maxlev=0, maxic=0;
  char ***string=NULL;
  char *dummy=NULL;
/*   int lc=0; */
  

  if (!input.quiet)
    fprintf (stderr, " ... reading cloud data from %s\n", input.filename[FN_WCLOUD_IPA]);
      
  if ((status = ASCII_checkfile (input.filename[FN_WCLOUD_IPA], &(output->nipa), 
				 &min_columns, &max_columns, &max_length)) != 0) {
    fprintf (stderr, " read_wc_ipa: could not open file %s\n", input.filename[FN_WCLOUD_IPA]);
    return status;
  }
  
  if (min_columns<2) {
    fprintf (stderr, "Error, need at least 2 columns in independent pixel file %s\n", 
	     input.filename[FN_WCLOUD_IPA]);
    return -1;
  }
  
  /* read ASCII file */
  if ((status = ASCII_calloc_string(&string, output->nipa, max_columns, FILENAME_MAX+1)) != 0) 
    return status;
  
  if ((status = ASCII_readfile(input.filename[FN_WCLOUD_IPA], string)) != 0) 
    return status;
  

  /* allocate memory for IPA arrays */
  output->wcipa     = calloc (output->nipa, sizeof(cld_out_struct));
  output->ipaweight = calloc (output->nipa, sizeof(float));
  
  /* now finally read the water cloud files */
  for (ipa=0; ipa<output->nipa; ipa++) {
    output->ipaweight[ipa] = (float) strtod (string[ipa][1], &dummy);
    
    if (!input.quiet)
      fprintf (stderr, " ... reading %s\n", string[ipa][0]);
    
    if (strlen(input.filename[FN_WCLOUD_PROP]) > 0) {

      status = read_cld_file (input, output->wl.nlambda_t,
			      string[ipa][0], &(output->wcipa[ipa]), 0, input.wc.layer, 
                              output->alt.altitude, output->atm.microphys.press,
                              output->atm.zd, output->atm.z_model_layer,
                              &output->cf, "wc");
      if (status!=0) {
	fprintf (stderr, "Error %d reading IPA %s\n", status, string[ipa][0]);
	return status;
      }
    }
    else {
      status = read_and_convert_wc_file (input, output,
					 string[ipa][0], &(output->wcipa[ipa]));
      if (status!=0) {
	fprintf (stderr, "Error %d reading IPA %s\n", status, string[ipa][0]);
	return status;
      }
    }
  }
  

  /* allocate memory for general wc structure where the single scattering properties */
  /* are stored, and where the independent columns are copied later                  */
  maxlev=0;
  maxic=0;
  for (ipa=0; ipa<output->nipa; ipa++) {
    if (maxlev < output->wcipa[ipa].nlev) {
      maxlev = output->wcipa[ipa].nlev;
      maxic = ipa;
    }
  }
  
  /* ????? */
  /* CE dirty fix here, for some reason nphamat is not set here */
  if( output->wc.optprop.nphamat==0)
    output->wc.optprop.nphamat=6;
  /* ????? */
  
  /* allocate memory */
  status = calloc_cld_out (&(output->wc), output->wl.nlambda_t, maxlev, 
                           output->wc.optprop.nphamat);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cld_out struct\n", status);
    return status;
  }


  /* Read optical properties file if available */
  if (strlen(input.filename[FN_WCLOUD_PROP]) > 0) {
    
    status = read_cloud_prop (input.filename[FN_WCLOUD_PROP], 
			      output->wl.lambda_t, output->wl.nlambda_t, 
			      output->wl.nlambda_rte_lower,
                              output->wl.nlambda_rte_upper, 
			      input.wc.interpolate, 0,
			      &(output->wc), input.wc.properties,
			      NULL, NULL, 0,
			      &(output->atm.nmom), input.rte.pol_nstokes,
			      input.rte.nstr,
			      input.rte.solver,
                              input.verbose, input.quiet);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading water cloud property file %s\n", 
	       status, input.filename[FN_WCLOUD_PROP]);
      return status;
    }

    if (input.verbose)
      fprintf (stderr, " ... read water cloud optical properties from %s\n", 
	       input.filename[FN_WCLOUD_PROP]);
  }

  /* copy independent pixel microphysical properties of the pixel with the most levels to wc->output */
  status = cp_cld_out (&(output->wc), output->wcipa[maxic], 0, 0, input.quiet);  /* in cloud.c */
  if (status!=0) {
    fprintf (stderr, "Error %d copying output->wcipa[ipa] to output->wc\n", status);
    return status;
  }

  if (input.verbose)
    fprintf (stderr, " ... read %d independent water cloud columns from %s\n", 
	     output->nipa, input.filename[FN_WCLOUD_IPA]);

/*   for (ipa=0; ipa<output->nipa; ipa++) { */
/*     fprintf (stderr, "cloud number %d\n", ipa+1); */
/*     for (lc=0 ; lc<output->wcipa[ipa].nlev ; lc++) */
/*       fprintf (stderr, "z= %6.2f, tau= %13.6e\n",output->wc.zd[lc],output->wcipa[ipa].optprop.dtau [0][lc]); */
/*   } */

  return 0;
}



int read_ic_ipa (input_struct input, output_struct *output)
{
  int ipa=0, min_columns=0, max_columns=0, max_length=0, status=0, maxlev=0, maxic=0;
  char ***string=NULL;
  char *dummy=NULL;

  float *tmpweight=NULL;
  int tmpn=0;
  

  if (!input.quiet)
    fprintf (stderr, " ... reading cloud data from %s\n", input.filename[FN_ICLOUD_IPA]);
  
  if ((status = ASCII_checkfile (input.filename[FN_ICLOUD_IPA], &tmpn, 
				 &min_columns, &max_columns, &max_length)) != 0) {
    fprintf(stderr,"read_ic_ipa: Could not open file %s\n", input.filename[FN_ICLOUD_IPA]);
    return status;
  }  

  if (min_columns<2) {
    fprintf (stderr, "Error, need at least 2 columns in independent pixel file %s\n", 
	     input.filename[FN_ICLOUD_IPA]);
    return -1;
  }
  

  /* check if water cloud and ice cloud files have same number of rows */
  if (input.wcipa && tmpn != output->nipa) {
    fprintf (stderr, "Error. %s and %s have different number of rows, %d and %d.\n",
	     input.filename[FN_ICLOUD_IPA], input.filename[FN_WCLOUD_IPA],
	     tmpn, output->nipa);
    return -1;
  }
  
  output->nipa = tmpn;
  

  /* read ASCII file */
  if ((status = ASCII_calloc_string(&string, output->nipa, max_columns, FILENAME_MAX+1)) != 0) 
    return status;
  
  if ((status = ASCII_readfile(input.filename[FN_ICLOUD_IPA], string)) != 0) 
    return status;
  

  /* allocate memory for IPA arrays */
  output->icipa     = calloc (output->nipa, sizeof(cld_out_struct));
  tmpweight = calloc (output->nipa, sizeof(float));
  

  /* extract weight */
  for (ipa=0; ipa<output->nipa; ipa++)
    tmpweight[ipa] = (float) strtod (string[ipa][1], &dummy);

  if (input.wcipa) {
    for (ipa=0; ipa<output->nipa; ipa++) {
      if (output->ipaweight[ipa] != tmpweight[ipa]) {
	fprintf (stderr, "Error, water and ice cloud weights are not identical:\n");
	fprintf (stderr, "       %f and %f in row %d\n", 
		 output->ipaweight[ipa], tmpweight[ipa], ipa);
	return -1;
      }
    }
    free(tmpweight);
  }
  else 
    output->ipaweight = tmpweight;
    
	

  /* now finally read the ice cloud files */
  for (ipa=0; ipa<output->nipa; ipa++) {
    if (!input.quiet)
      fprintf (stderr, " ... reading %s\n", string[ipa][0]);
    
    if (strlen(input.filename[FN_ICLOUD_PROP]) > 0) {

      status = read_cld_file (input, output->wl.nlambda_t,
			      string[ipa][0], &(output->icipa[ipa]), 0, input.ic.layer, 
                              output->alt.altitude, output->atm.microphys.press,
                              output->atm.zd, output->atm.z_model_layer,
                              &output->cf, "ic");          /* in cloud.c */
      if (status!=0) {
	fprintf (stderr, "Error %d reading IPA %s\n", status, string[ipa][0]);
	return status;
      }
    }
    else {
      status = read_and_convert_ic_file (input, output,
					  string[ipa][0], &(output->icipa[ipa]));
      if (status!=0) {
	fprintf (stderr, "Error %d reading IPA %s\n", status, string[ipa][0]);
	return status;
      }
    }
  }

  /* allocate memory for general ic structure where the single scattering properties */
  /* are stored, and where the independent columns are copied later                  */
  maxlev=0;
  maxic=0;
  for (ipa=0; ipa<output->nipa; ipa++) 
    if (maxlev < output->icipa[ipa].nlev) {
      maxlev = output->icipa[ipa].nlev;
      maxic = ipa;
    }
  
  /* ????? */
  /* CE dirty fix here, for some reason nphamat is not set here */
  if( output->ic.optprop.nphamat==0)
    output->ic.optprop.nphamat=6;
  /* ????? */
  

  /* allocate memory */
  status = calloc_cld_out (&(output->ic), output->wl.nlambda_t, maxlev, output->ic.optprop.nphamat);
  if (status!=0) {
    fprintf (stderr, "Error %d allocating memory for cld_out struct (line %d, function %s in %s)\n", 
                     status, __LINE__, __func__, __FILE__);
    return status;
  }


  /* Read optical properties file if available */
  if (strlen(input.filename[FN_ICLOUD_PROP]) > 0) {
    
    status = read_cloud_prop (input.filename[FN_ICLOUD_PROP], 
			      output->wl.lambda_t, output->wl.nlambda_t, 
			      output->wl.nlambda_rte_lower, 
                              output->wl.nlambda_rte_upper, 
			      input.ic.interpolate, 0,
			      &(output->ic), input.ic.properties,
			      NULL, NULL, 0,
			      &(output->atm.nmom), input.rte.pol_nstokes,
			      input.rte.nstr,
			      input.rte.solver,
                              input.verbose, input.quiet);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading ice cloud property file %s\n", 
		 status, input.filename[FN_ICLOUD_PROP]);
      return status;
    }

    if (input.verbose)
      fprintf (stderr, " ... read ice cloud optical properties from %s\n", 
	       input.filename[FN_ICLOUD_PROP]);
  }

  /* copy independent pixel microphysical properties of the pixel with the most levels to ic->output */
  status = cp_cld_out (&(output->ic), output->icipa[maxic], 0, 0, input.quiet);
  if (status!=0) {
    fprintf (stderr, "Error %d copying output->icipa[ipa] to output->ic\n", status);
    return status;
  }

  if (input.verbose)
    fprintf (stderr, " ... read %d independent ice cloud columns from %s\n", 
	     output->nipa, input.filename[FN_ICLOUD_IPA]);

  return 0;
}
